"""
Provide mixins for throttling/rate limiting in views.

Depends on ``django-axes``.
"""

from collections.abc import Container
from time import time
from typing import Literal

from django.core.cache import caches
from django.core.exceptions import ImproperlyConfigured, PermissionDenied
from django.http import HttpRequest

from axes.helpers import get_client_ip_address


class ThrottleMixin:
    """
    A very simple throttling implementation with, hopefully, sane defaults.

    You can specifiy the amount of visits (throttle_visits) a view can get,
    for a specific period (in seconds) throttle_period.
    """

    # n visits per period (in seconds)
    throttle_visits = 100
    throttle_period = 60**2  # in seconds
    throttle_403 = True
    throttle_name = "default"

    # get and options should always be fast. By default
    # do not throttle them.
    throttle_methods: Container[str] | Literal["all"] = (
        "post",
        "put",
        "patch",
        "delete",
        "head",
        "trace",
    )

    request: HttpRequest

    def get_throttle_cache(self):
        return caches["default"]

    def get_throttle_identifier(self) -> str:
        user = getattr(self, "user_cache", self.request.user)
        return str(user.pk)

    def create_throttle_key(self):
        """
        :rtype string Use as key to save the last access
        """

        return (
            f"throttling_{self.get_throttle_identifier()}_"
            f"{self.throttle_name}_{self.get_throttle_window()}"
        )

    def get_throttle_window(self):
        """
        round down to the throttle_period, which is then used to create the key.
        """
        current_time = int(time())
        return current_time - (current_time % self.throttle_period)

    def get_visits_in_window(self):
        cache = self.get_throttle_cache()
        key = self.create_throttle_key()

        initial_visits = 1
        stored = cache.add(key, initial_visits, self.throttle_period)
        if stored:
            visits = initial_visits
        else:
            try:
                visits = cache.incr(key)
            except ValueError:
                visits = initial_visits
        return visits

    def should_be_throttled(self):
        if self.throttle_methods == "all":
            return True
        assert isinstance(self.request.method, str)
        return self.request.method.lower() in self.throttle_methods

    def dispatch(self, request, *args, **kwargs):
        if self.throttle_403:
            if (
                self.should_be_throttled()
                and self.get_visits_in_window() > self.throttle_visits
            ):
                raise PermissionDenied

        return super().dispatch(request, *args, **kwargs)  # pyright:ignore[reportAttributeAccessIssue]


class IPThrottleMixin(ThrottleMixin):
    """
    Same behavior as ThrottleMixin except it limits the amount of tries
    per IP-address a user can access a certain view.
    """

    def get_throttle_identifier(self):
        ip_address = get_client_ip_address(self.request)
        if ip_address is None:
            raise ImproperlyConfigured(
                "Could not determine IP address. "
                "Check your reverse proxy configuration."
            )
        return ip_address



"""
Combine maykin-2fa and django-admin-index utilities.

Depends on ``django-admin-index``.
"""

from django_admin_index.utils import (
    should_display_dropdown_menu as default_should_display_dropdown_menu,
)


def should_display_dropdown_menu(request) -> bool:
    """
    Overrides the default `should_display_dropdown_menu` from `django_admin_index.utils`
    to:
    - not display the dropdown in `maykin_2fa` views.
    - only display the dropdown to verified users.
    """
    default = default_should_display_dropdown_menu(request)

    # never display the dropdown in two-factor admin views
    if request.resolver_match.view_name.startswith("maykin_2fa:"):
        return False

    # do not display the dropdown until the user is verified.
    return default and request.user.is_verified()



"""
Custom Django system checks to prevent common mistakes.
"""

import os

from django.conf import settings
from django.core.checks import Warning, register


@register
def check_missing_init_files(app_configs, **kwargs):
    """
    Check that all packages have __init__.py files.

    If they don't, the code will still run, but tests aren't picked up by the
    test runner, for example.
    """
    errors = []

    for dirpath, _, filenames in os.walk(settings.DJANGO_PROJECT_DIR):
        dirname = os.path.split(dirpath)[1]
        if dirname == "__pycache__":
            continue

        if "__init__.py" in filenames:
            continue

        extensions = [os.path.splitext(fn)[1] for fn in filenames]
        if ".py" not in extensions:
            continue

        errors.append(
            Warning(
                f"Directory {dirpath} does not contain an `__init__.py` file {dirname}",
                hint="Consider adding this module to make sure tests are picked up",
                id="maykin.W001",
            )
        )

    return errors



from django.conf import settings

GOOGLE_ANALYTICS_ID = None
"""
Google analytics id added to the request context.
"""

SHOW_ENVIRONMENT = False
"""
Enable or disable the ``{% environment_info %}`` template tag.
"""

ENVIRONMENT_BACKGROUND_COLOR = "orange"
"""
The background color of the ``{% environment_info %}`` template tag.
"""

ENVIRONMENT_FOREGROUND_COLOR = "black"
"""
The foreground color of the ``{% environment_info %}`` template tag.
"""

ENVIRONMENT_LABEL = None
"""
The textual content of the ``{% environment_info %}`` template tag.
"""

RELEASE = None
"""
The release version shown in the ``{% version_info %}`` template tag.
"""

GIT_SHA = None
"""
The commit hash shown in the ``{% version_info %}`` template tag.
"""

PDF_BASE_URL_FUNCTION = None
"""
Function that returns the base url needed to download/resolve custom fonts and/or any
image URLs included in the document to render.

Required for the :ref:`quickstart_pdf` extra.
"""


def get_setting(name: str):
    default = globals()[name]
    return getattr(settings, name, default)



from django import http
from django.template import TemplateDoesNotExist, loader
from django.views.decorators.csrf import requires_csrf_token
from django.views.defaults import ERROR_500_TEMPLATE_NAME


@requires_csrf_token
def server_error(
    request, template_name=ERROR_500_TEMPLATE_NAME
) -> http.HttpResponseServerError:
    """
    500 error handler.
    """

    try:
        template = loader.get_template(template_name)
    except TemplateDoesNotExist:
        if template_name != ERROR_500_TEMPLATE_NAME:
            # Reraise if it's a missing custom template.
            raise
        return http.HttpResponseServerError(
            b"<h1>Server Error (500)</h1>", content_type="text/html"
        )
    context = {"request": request}
    return http.HttpResponseServerError(template.render(context))
